//
// Binary tree predictive encoder
//
// Converts raw PPM and PGM format files to BTPC encoded files using the
// function btpc_encode()
//
// Version 5.
//
// Copyright (c) John A Robinson. 1994, 1995, 1997, 2002, 2003
//
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "btpc.h"

//
// Helper functions for dealing with PPM/PGM files
//

void pgmword(FILE *in, char *cbuf)
// The PNM man pages say you have to be as lenient as possible about pgm/ppm/pbm
// files, so this is my general-purpose lenient word reader.
// Read next PGM/PPM/PBM word, ignoring comments and whitespace
	{
	char c;
	while(1) {
	    do 
		c = fgetc(in);
	    while(c == ' ' || c == '\t' || c == '\n' || c == '\r');
	    if (c == '#') {
		do
			c = fgetc(in);
		while (c != '\n');
		}
	    else break;
	    }
	*cbuf++ = c;
	c = fgetc(in);
	while (c != ' ' && c != '\t' && c != '\n' && c != '\r' && c != '#')
	    {
	    *cbuf++ = c;
	    c = fgetc(in);
	    }
	fseek(in, -1, SEEK_CUR);
	*cbuf = '\0';
	}

int readpnmandpad(const char *pgmname, int& rows, int& cols, 
	int& ymax, int& xmax,
	unsigned char ***pbuf, int numlevels)
// Read picture in "P4", "P5" or "P6" format into arrays. Pad the picture 
// size up to the next multiple of blocksize.
	{
	int i,j;
	int maxval;
	char cbuf[32];
	unsigned char c;
	int retval = 0;
	FILE *in=fopen(pgmname, "rb");
	if (!in) {
		fprintf(stderr, "Cannot open %s\n", pgmname);
		return -1;
		}
	cbuf[0] = fgetc(in);
	cbuf[1] = fgetc(in);
	cbuf[2] = 0;
	if (!strcmp(cbuf,"P5")) 
		// PGM image
		retval = 1;
	else if (!strcmp(cbuf,"P4"))
		// PBM image
		retval = 0;
	else if (!strcmp(cbuf,"P6"))
		// PPM image
		retval = 3;
	else {
		fprintf(stderr,"Sorry. This version of cbtpc only works with ");
		fprintf(stderr,"raw format (P4, P5 or P6) PBM/PGM/PPM files.\n");
		fprintf(stderr,"%s is not a valid raw format file\n",pgmname);
		return -1;
		}
	pgmword(in, cbuf);
	cols = atoi(cbuf);
	pgmword(in,cbuf);
	rows = atoi(cbuf);
	if (retval) {
		pgmword(in,cbuf);
		maxval = atoi(cbuf);
		// We're going to throw this maxval away anyhow
		}
	// absorb final whitespace char before start of data
	c = fgetc(in);
	// Make ymax equal to next highest multiple of blocksize
	ymax = ((rows+(1<<numlevels)-1)>>numlevels)<<numlevels;
	// Make xmax equal to next highest multiple of 2*blocksize
	// Do this so that points can be processed in pairs in binary tree
	xmax = ((cols+(1<<(numlevels+1))-1)>>(numlevels+1))<<(numlevels+1);
	if (retval == 0)	// PBM: We map white to 255 and black to 0.
				// Later processing in encode detects bilevel
	    {
	    unsigned char **buf;
	    unsigned char *tempbuf = new unsigned char[(cols+7)/8];
	    *pbuf++ = buf = new unsigned char *[ymax];
	    *pbuf++ = buf;
	    *pbuf = buf;	// So all three components point at same
	    for (i = 0; i < rows; i++) {
		buf[i] = new unsigned char[xmax];
		fread(tempbuf, 1, (cols+7)/8, in);
		unsigned char *tb = tempbuf;
		int bitmask = 0x80;
		for (j = 0; j < cols; j++) {
			buf[i][j] = (*tb & bitmask) ? 0 : 255;
			bitmask >>= 1;
			if (!bitmask) {
				bitmask = 0x80;
				tb++;
				}
			}
		for (j = cols; j < xmax; j++)	// Pad image by
			buf[i][j] = buf[i][cols-1];	// copying last pel
		}
	    // Vertical padding
	    for (i = rows; i < ymax; i++) {
		buf[i] = new unsigned char[xmax];	// Pad image by
		memcpy(buf[i], buf[rows-1], xmax);	// copying last line
		}
	    delete [] tempbuf;
	    }
	else if (retval == 1)	// PGM straightforward
	    {
	    unsigned char **buf;
	    *pbuf++ = buf = new unsigned char *[ymax];
	    *pbuf++ = buf;
	    *pbuf = buf;	// So all three components point at same
	    for (i = 0; i < rows; i++) {
		buf[i] = new unsigned char[xmax];
		fread(buf[i], 1, cols, in);
		for (j = cols; j < xmax; j++)	// Pad image by
			buf[i][j] = buf[i][cols-1];	// copying last pel
		}
	    // Vertical padding
	    for (i = rows; i < ymax; i++) {
		buf[i] = new unsigned char[xmax];	// Pad image by
		memcpy(buf[i], buf[rows-1], xmax);	// copying last line
		}
	    }
	else			// PPM
	    {
	    unsigned char **bufr, **bufg, **bufb;
	    unsigned char *tempbuf = new unsigned char[3*xmax];
	    unsigned char *pr, *pg, *pb, *ptemp;
	    *pbuf++ = bufg = new unsigned char *[ymax];
	    *pbuf++ = bufr = new unsigned char *[ymax];
	    *pbuf = bufb = new unsigned char *[ymax];
	    // This loop incorporates vertical padding
	    for (i = 0; i < ymax; i++){
		pg = bufg[i] = new unsigned char[xmax];
		pr = bufr[i] = new unsigned char[xmax];
		pb = bufb[i] = new unsigned char[xmax];
		if (i < rows)
			fread(tempbuf,1,cols*3,in);
		ptemp = tempbuf;
		for (j = 0; j < cols; j++) {
			*pr++ = *ptemp++;
			*pg++ = *ptemp++;
			*pb++ = *ptemp++;
			}
		for (j = cols; j < xmax; j++) {	// Pad
			*pr++ = bufr[i][cols-1];
			*pg++ = bufg[i][cols-1];
			*pb++ = bufb[i][cols-1];
			}
		}
	    delete [] tempbuf;
	    }
	fclose(in);
	return retval;
	}

int main(int argc, const char *argv[])
	{
	if ((argc != 3)&&(argc != 4))
		{
		fprintf(stderr,"Usage: cbtpc pnm_inpic code_file [quality]\n");
		return -1;
		}
	unsigned char **components[3];
	// char comment[] = "pnm to btpc by cbtpc5.1";
	char comment[] = "";
	int truerows, truecols;
	int padrows, padcols;
	int pnmtype;
	int numlevels = btpc_get_numlevels();
	pnmtype = readpnmandpad(argv[1],truerows,truecols,
		padrows, padcols,components, numlevels);
	if (pnmtype < 0)
		return -1;
	int quality = 75;
	if (argc == 4)
		quality = atoi(argv[3]);
	if (pnmtype == 0) {
		quality = 100;	// Always code at full quality
		pnmtype = 1;	// Just represent as PGM now
		}
	int retval;
	FILE *cfp = fopen(argv[2],"wb");
	if (cfp == 0) 
		retval = -2;
	else
		retval = btpc_encode(components, cfp, truerows, truecols,
		padrows, padcols, pnmtype, quality, comment);
	switch (retval) {
		case -1:
			fprintf(stderr,"Quality parameter must be between 0 and 100\n");
			break;
		case -2:
			fprintf(stderr,"Can't open %s\n", argv[2]);
			break;
		default:
			// Success
			break;
		}
	for (int compnum = 0; compnum < pnmtype; compnum++) {
		unsigned char **in = components[compnum];
	   	for (int i = 0; i < padrows; i++)
			delete [] in[i];
	    	delete [] in;
	    	}
	return retval;
	}

