// New colour conversion transforms for BTPC using PCA
// John Robinson 2002.
#include <stdio.h>
#include <math.h>
#ifndef M_PI
#define M_PI		3.14159265358979323846	/* pi */
#endif
const int shiftbits = 15;

// Simplified PCA class for 3D PCA only
class pca3d {
	int numsamples;
	double totals[3];
	double crosstotals[3][3];
	double stddevs[3];
	double mean[3];
	double cov[3][3];
	void calc_mean_and_covariance() {
		int i, j;
		for (i = 0; i < 3; i++) {
			mean[i] = totals[i]/numsamples;
			}
		for (i = 0; i < 3; i++) {
			for (j = i; j < 3; j++) {
				cov[i][j] = crosstotals[i][j]/numsamples
					 - mean[i]*mean[j];
				if (i != j)
					cov[j][i] = cov[i][j];
				}
			}
		}
public:
	pca3d() {
		for (int i = 0; i < 3; i++) {
			totals[i] = 0.0;
			crosstotals[i][0] = 0.0;
			crosstotals[i][1] = 0.0;
			crosstotals[i][2] = 0.0;
			}
		numsamples = 0;
		}
	~pca3d() {
		}

	void addsample(int *sample) {
		int i, j;
		for (i = 0; i < 3; i++) {
			totals[i] += (double) sample[i];
			for (j = i; j < 3; j++)
				crosstotals[i][j] += (double) sample[i]*sample[j];
			}
		numsamples++;
		}
	
	void getpcs(double *mout, double *eigval, double *eigvec) {
		calc_mean_and_covariance();
		// Get eigenvalues
		// Set up cubic equation from determinant of cov mx
		double a = -(cov[0][0] + cov[1][1] + cov[2][2]);
		double b = (cov[1][1]*cov[2][2]+cov[0][0]*cov[1][1]
			+cov[0][0]*cov[2][2] -cov[1][2]*cov[1][2]
			-cov[0][1]*cov[0][1]-cov[0][2]*cov[0][2]);
		double c = (cov[0][0]*cov[1][2]*cov[1][2] 
			+cov[2][2]*cov[0][1]*cov[0][1] 
			+cov[1][1]*cov[0][2]*cov[0][2] 
			-cov[0][0]*cov[1][1]*cov[2][2]
			-2*cov[0][1]*cov[0][2]*cov[1][2]);
		double p = (3*b-a*a)/3.0;
		double q = c + 2*a*a*a/27.0 - a*b/3.0;
		// Assume irreducible case, so
		double temp = fabs(p)/3.0;
		double costheta = -q/(2*sqrt(temp*temp*temp));
		double theta = acos(costheta);
		temp = sqrt(temp);
		eigval[0] = -a/3.0+2*temp*cos(theta/3.0);
		eigval[1] = -a/3.0-2*temp*cos((theta-M_PI)/3.0);
		eigval[2] = -a/3.0-2*temp*cos((theta+M_PI)/3.0);
		// Bubble sort: Desceding order
		if (eigval[0] < eigval[1]) {
		    temp=eigval[1];eigval[1]=eigval[0];eigval[0]=temp;
		    }
		if (eigval[1] < eigval[2]) {
		    temp=eigval[2];eigval[2]=eigval[1];eigval[1]=temp;
		    }
		if (eigval[0] < eigval[1]) {
		    temp=eigval[1];eigval[1]=eigval[0];eigval[0]=temp;
		    }
		int i;
		for (i = 0; i < 3; i++) {
		    // First solve equations as if x1 = 1
		    double x2, x3;
		    x2 = (cov[0][2]*cov[0][1]-cov[1][2]*(cov[0][0]-eigval[i]));
		    x2 /= (cov[1][2]*cov[0][1]-cov[0][2]*(cov[1][1]-eigval[i]));
		    x3 = (cov[0][1]*cov[0][1]-(cov[1][1]-eigval[i])*
						(cov[0][0]-eigval[i]));
		    x3 /= (cov[1][1]-eigval[i])*cov[0][2] - cov[0][1]*cov[1][2];
		    // And normalize
		    double total = 1 + x2*x2 + x3*x3;
		    temp = 1/sqrt(total);
		    eigvec[3*i] = temp;
		    eigvec[3*i+1] = x2*temp;
		    eigvec[3*i+2] = x3*temp;
		    }
		   
		for (i = 0; i < 3; i++) 
			mout[i] = mean[i];
		}
	};


void calc_colour_transform(unsigned char ***components, int *transform, 
		int num_rows, int num_cols) {
  	register unsigned char *pg;
  	register unsigned char *pr;
  	register unsigned char *pb;
  	int colvals[3];
  	pca3d dopca;
  	for (int row = 0; row < num_rows; row++) {
    		pg = components[0][row];
    		pr = components[1][row];
    		pb = components[2][row];
    		for (int col = 0; col < num_cols; col++) {
      			colvals[0] = *pg++;
      			colvals[1] = *pr++;
      			colvals[2] = *pb++;
      			dopca.addsample(colvals);
			}
		}
	double means[3];
  	double eigval[3];
  	double eigvec[3][3];
	dopca.getpcs(means, eigval, (double *)eigvec);
	// Ensure first e vector is positive
	if (eigvec[0][0] < 0) {
		for (int j = 0; j < 3; j++)
			eigvec[0][j] = -eigvec[0][j];
		}
  	for (int i = 0; i < 3; i++) {
		// Now change to ints
		for (int j = 0; j < 3; j++)
			*transform++ =  (int)((1<<shiftbits)*eigvec[i][j]);
		}
	}
void rgb_pca(unsigned char ***components, int *transform, 
		int num_rows, int num_cols) {
	register int r, g, b;
  	register unsigned char *pg;
  	register unsigned char *pr;
  	register unsigned char *pb;
	double t[9];
	for (int i = 0; i < 9; i++)
		t[i] = ((double)transform[i])/(1<<shiftbits);
  	for (int row = 0; row < num_rows; row++) {
    		pg = components[0][row];
    		pr = components[1][row];
    		pb = components[2][row];
    		for (int col = 0; col < num_cols; col++) {
      			g = *pg;
      			r = *pr;
      			b = *pb;
			*pg++ = (unsigned char) ((t[0]*g + t[1]*r + t[2]*b)/2 + 0.5);
			*pr++ = (unsigned char) ((t[3]*g + t[4]*r + t[5]*b)/2 + 128.5);
			*pb++ = (unsigned char) ((t[6]*g + t[7]*r + t[8]*b)/2 + 128.5);
			}
		}
	}
void pca_rgb(unsigned char ***components, int *transform, 
		int num_rows, int num_cols, const unsigned char **clip) {
	register int r, g, b;
  	register unsigned char *pg;
  	register unsigned char *pr;
  	register unsigned char *pb;
	const unsigned char *clip0 = clip[0];
	const unsigned char *clip1 = clip[1];
	const unsigned char *clip2 = clip[2];
	long int g_to_g[256];
	long int g_to_r[256];
	long int g_to_b[256];
	long int r_to_g[256];
	long int r_to_r[256];
	long int r_to_b[256];
	long int b_to_g[256];
	long int b_to_r[256];
	long int b_to_b[256];
// Set up lookup tables
	for (int i = 0; i < 256; i++) {
		g_to_g[i] = transform[0]*2*i;
		g_to_r[i] = transform[1]*2*i;
		g_to_b[i] = transform[2]*2*i;
		r_to_g[i] = transform[3]*2*(i-128);
		r_to_r[i] = transform[4]*2*(i-128);
		r_to_b[i] = transform[5]*2*(i-128);
		b_to_g[i] = transform[6]*2*(i-128);
		b_to_r[i] = transform[7]*2*(i-128);
		b_to_b[i] = transform[8]*2*(i-128);
		}
  	for (int row = 0; row < num_rows; row++) {
    		pg = components[0][row];
    		pr = components[1][row];
    		pb = components[2][row];
    		for (int col = 0; col < num_cols; col++) {
      			g = *pg;
      			r = *pr;
      			b = *pb;
			*pg++=clip0[(g_to_g[g]+r_to_g[r]+b_to_g[b]+(1<<(shiftbits-1)))>>shiftbits];
			*pr++=clip1[(g_to_r[g]+r_to_r[r]+b_to_r[b]+(1<<(shiftbits-1)))>>shiftbits];
			*pb++=clip2[(g_to_b[g]+r_to_b[r]+b_to_b[b]+(1<<(shiftbits-1)))>>shiftbits];
			}
		}
	}

