//
//
// Binary Tree Predictive Coding.
// John Robinson, 1994, 1995, 1997, 2002, 2003.
//
// Header file for static huffman coder and flagbits coder.
//

#include <stdio.h>
// #define DEBUG 1
// #define NOCOMPACTORDEBUG 1

const int TABLEBITS = 12;
const int MASK = (1<<TABLEBITS)-1;
const int TABLESIZE = (1<<TABLEBITS);

class compactor {
private:
	int debug;

	int id;
	class cfileio *pfio;
	int longest;
	int longmask;
	unsigned char num_symbols_of_length[32];
	unsigned char ordered_symbols[256];
	long savefreq[256];
	long accum;
	int maxval;
	int litlength;
	int usinghuffman;	// 1 if using Huffman, 0 if literals
// The shared data area:
	union {
		struct {
			long f[256];
			unsigned short enc[256];
			unsigned char len[256];
			} enctables;
		struct {
			unsigned char dec[TABLESIZE];
			unsigned char len[TABLESIZE];
			} dectables;
		} shared;
// Next the stuff for encoding
	long *freq;
	unsigned short *encode;
	unsigned char *enclength;
// Then the stuff for decoding
	unsigned char *decode;
	unsigned char *declength;
	int usetree();

public:
	compactor(){
		debug = 0;
		id = 0;
		freq = shared.enctables.f;
		encode = shared.enctables.enc;
		enclength = shared.enctables.len;
		decode = shared.dectables.dec;
		declength = shared.dectables.len;
		pfio = 0; 
		longest = 0;
		reset();
		}
	~compactor() {}
	void putbit(int val);
	int getbit();
	void reset();
	void debugon() { debug = 1; }
	void debugoff() { debug = 0; }
	void indexnumber(int i) { id = i; }
	void attach(class cfileio *f) { pfio = f; }
	void inc(int symbol) {freq[symbol]++;};
	void dec(int symbol) {freq[symbol]--;};
	void buildtree(void);
	void writetree(void);
	void maketable_encode(void);
	void put(int inbyte);
	void incorput(int symbol, int which) {
		if (which)
			put(symbol);
		else
			freq[symbol]++;
		}
	void readtree(void);
	void maketable_decode(void);
	int get();
	void copytree(compactor& source);
	int copytreewithout(compactor& source, int from, int to);
	};

class cfileio {
private:
	FILE	*cfp;
// Output works a byte at a time and is therefore simple
	int	outbyte;
	int	outbits; // Number of output bits
// Input is double buffered: first get a chunk into buf, then get bits into
// bitbuf
	int charsleft;
	unsigned long	bitbuf;
	int valid;	// bits left in bitbuf
	unsigned char *pbuf;
	unsigned char buf[2048];
	int fillbuf() {
		return(charsleft = fread((pbuf = buf),1,2048,cfp));
		}
	void loadforlook() {
		bitbuf = (bitbuf<<8) + (charsleft-- ? *pbuf++ : 
			(fillbuf() && charsleft--) ? *pbuf++ : 0);
		}
	int lookbits(const int& bits, const int& mask) {
		if (valid < bits)
			{
			loadforlook();
			valid += 8;
			}
		if (valid < bits)
			{
			loadforlook();
			valid += 8;
			}
		return (int) ((bitbuf >> (valid-bits)) & mask);
		}
	void skipbits(const int bits) {
		valid -= bits;
		}
	int get(unsigned char *declength, unsigned char *decode, 
	    register int bits, const int mask) {
		register int v = valid;
		if (v < bits) {
			loadforlook();
			if ((v += 8) < bits) {
				loadforlook();
				v += 8;
				}
			}
		bits = (int) ((bitbuf >> (v-bits)) & mask);
		valid = v - declength[bits];
		return decode[bits];
		}
public:
	cfileio(FILE *file) { 
		cfp = file; outbyte = 0; outbits = 0;
		bitbuf = 0; charsleft = 0; valid = 0; }
	void putbit(int val) {
		outbyte <<= 1;
		if (val)
			outbyte++;
		outbits++;
		if ((outbits &= 0x7) == 0)
			putc(outbyte, cfp);
		}
	void putbits(int val, int numbits) {
		int space = 8-outbits;
		if (numbits > space) {
			outbyte <<= space;
			outbyte |= val>>(numbits = numbits-space);
			putc(outbyte, cfp);
			outbits = 0;
			space = 8;
			val &= (1<<numbits) - 1;
			if (numbits > space) { 
			// Now space = 8 and outbyte is empty
			    outbyte = val>>(numbits = numbits-space);
			    putc(outbyte, cfp);
			    val &= (1<<numbits) - 1;
			    }
			}
		// Now guaranteed that numbits <= space
		outbyte <<= numbits;
		outbyte |= val;
		outbits += numbits;
		if ((outbits &= 0x7) == 0)
			putc(outbyte, cfp);
		}
	void putbyte(unsigned char val) {
		int thisbyte = 8 - outbits;
		outbyte <<= thisbyte;
		outbyte |= (val >> outbits);
		putc(outbyte, cfp);
		outbyte = val;
		}
	void flush() {
		if (outbits) {
			outbyte <<= 8 - outbits;
			putc(outbyte ,cfp);
			outbits = 0;
			}
		fflush(cfp);
		}
	int getbit() {
		if (!valid) {
			bitbuf = (charsleft-- ? *pbuf++ : 
				(fillbuf() && charsleft--) ? *pbuf++ : 0);
			valid = 8;
			}
		return (int) ((bitbuf >> --valid) & 1);
		}
	int getbits(int numbits) {
		// numbits <= 8
		int mask = (1<<numbits) - 1;
		int retval = lookbits(numbits, mask);
		skipbits(numbits);
		return retval;
		}
	~cfileio() {}
friend class compactor;
	};

inline void compactor::putbit(int val) {
#ifdef DEBUG
	if (debug)
		printf("Coder %d: bit %d\n",id,val);
#endif
	pfio->putbit(val);
	}
inline int compactor::getbit() {
	int code = pfio->getbit();
#ifdef DEBUG
	if (debug)
		printf("Coder %d: bit %d\n",id,code);
#endif
	return code;
	}
inline int compactor::get ()
{
	int code;
	if (!usinghuffman) {
		code = pfio->getbits(litlength);
#ifdef DEBUG
	if (debug) {
		printf("Coder %d: uncoded symbol %d (0x%x)\n",id,code,code);
		fflush(stdout);
		}
#endif
		return code;
		}
	code = pfio->get(declength, decode, longest, longmask);
#ifdef DEBUG
	if (debug) {
		printf("Coder %d: symbol %d (0x%x)\n",id,code,code);
		fflush(stdout);
		}
#endif
	return code;
}

const int FLAGWORDLENGTH = 4;
class rlcomp {
private:
	int debug;
	int id;
	int usingflagwords;	// Flag to say whether using flagwords
	int symbolcnt;
	int doublecode;		// Two successive flagwords of all freqsymbol
	unsigned char freqsymbol;
	unsigned char decodeflagword;
	unsigned char *inbuf;
	unsigned char *procbuf;
	unsigned char *readproc;
	unsigned char *inbufend;
	unsigned char *procbufend;
	int buflength;
	compactor *psymbols;
	compactor *pflagwords;
	long freq[256];
	void disable_compactor_debug() {
		if (!debug)
			return;
		psymbols->debugoff();
		pflagwords->debugoff();
		}
	void enable_compactor_debug() {
		if (!debug)
			return;
		psymbols->debugon();
		pflagwords->debugon();
		}
public:
	rlcomp();
	~rlcomp();
	int encodelength(int maxlength); // Called to set up encode buffers
	void putbit(int val) { psymbols->putbit(val);}
	int getbit() { return psymbols->getbit();}
	void reset();
	void debugon() { 
		debug=1; psymbols->debugon(); pflagwords->debugon();}
	void debugoff() { 
		debug=0; psymbols->debugoff(); pflagwords->debugoff();}
	void indexnumber(int i) { 
		id = i;
		psymbols->indexnumber(i+100); 
		pflagwords->indexnumber(i+200);
		}
	void attach(class cfileio *f) { 
		psymbols->attach(f); pflagwords->attach(f);
		}
	void inc(int symbol) {
		freq[symbol]++; 
		*inbufend++ = (unsigned char) symbol;
		}
	void buildtree(void);
	void writetree(void);
	void readtree(void);
	void maketable_encode(void);
	void maketable_decode(void);
	void incorput(int symbol, int which);
	void put(int inbyte);
	int get(){
		if (!usingflagwords)
			return psymbols->get();
		if (!symbolcnt) {
			if (doublecode) {
				doublecode--;
				decodeflagword = (1<<FLAGWORDLENGTH) - 1;
				}
			else {
				decodeflagword = pflagwords->get();
				if (decodeflagword == (1<<FLAGWORDLENGTH)) {
					decodeflagword--;
					doublecode = 1;
					}
				}
			}
		else
			decodeflagword <<= 1;
		if (++symbolcnt == FLAGWORDLENGTH)
			symbolcnt = 0;
		if (decodeflagword & (1<<(FLAGWORDLENGTH-1)))
			return freqsymbol;
		return psymbols->get();
		}
	};

const int numcoders = 8;
const int SST = 0; // First numbered coder for square band
const int DST = 4; // First numbered coder for diamond band
