//
// Binary tree predictive coding
// Helper routines for decoding, plus main decoding function
//
// Copyright (C) John Robinson, 1994, 1995, 1997, 2002, 2003.
//
#include "string.h"
#include "btpc.h"
#include "order.h"
// #define DEBUG 1

// Following table is used to clip
// Larger than 512 now, because color clipping can require more
static unsigned char Aclip[3][512+256];
static unsigned char *clip[3];

const unsigned char **initclip(const int *minval, const int *maxval,
	int pnmtype) {
    int i;
    for (int comp = 0; comp < pnmtype; comp++) {
    	clip[comp] = &Aclip[comp][256];
    	for (i = -256; i < minval[comp]; i++)
		clip[comp][i] = minval[comp];
    	for (i = minval[comp]; i < maxval[comp]; i++)
		clip[comp][i] = i;
    	for (; i < 512; i++)
		clip[comp][i] = maxval[comp];
	}
    return (const unsigned char **) clip;
}


static int Asq_decode_table[513] = { 0};
static int *sq_decode_table;
static int sq_range;
static int sq_out_of_range[256];
static int Adi_decode_table[513] = { 0};
static int *di_decode_table;
static int di_range;
static int di_out_of_range[256];
static int Aabt[513];
static int *abt;

int prepare_decoder(const int& sq_quant_spacing, const int& di_quant_spacing,
	const int& maxval)
// Sets up tables of representative levels to use in the current bands.
{
    int i;
    abt = &Aabt[256];
    abt[0] = 0;
    for (i = 1; i < 256; i++) {
	abt[i] = i;
	abt[-i] = i;
	}
    // Following four lines required to initialize tables if btpc_decode()
    // used repeatedly in same program. Previously relied on the
    // static initializations above.
    for (i = 0; i < 513; i++) {
	Asq_decode_table[i] = 0;
	Adi_decode_table[i] = 0;
	}
    // First do the square band
    sq_decode_table = &Asq_decode_table[256];
    int rep_level = sq_quant_spacing + (sq_quant_spacing-1)/2;
    if (rep_level != sq_decode_table[1]){	// Already loaded
	sq_range = maxval/sq_quant_spacing + 1;
	sq_decode_table[0] = 0;
	for (i = 1; i <= sq_range; i++, rep_level += sq_quant_spacing) {
	    sq_decode_table[i] = rep_level;
	    sq_decode_table[-i] = -rep_level;
	}
	for (i = 0; i < 256; i++)
	    {
	    sq_out_of_range[i] = (maxval-i)/sq_quant_spacing;
	    }
    }
    // Then the diamond band
    di_decode_table = &Adi_decode_table[256];
    rep_level = di_quant_spacing + (di_quant_spacing-1)/2;
    if (rep_level == di_decode_table[1])	// Already loaded
    	return 0;
    di_range = maxval/di_quant_spacing + 1;
    di_decode_table[0] = 0;
    for (i = 1; i <= di_range; i++, rep_level += di_quant_spacing) {
	di_decode_table[i] = rep_level;
	di_decode_table[-i] = -rep_level;
    }
    for (i = 0; i < 256; i++)
	{
	di_out_of_range[i] = (maxval-i)/di_quant_spacing;
	}
    return 0;
}


static char commentbuf[256];
int btpc_read_header(FILE *cfp, int& pnmtype, int& rows, int& cols, 
int *minval, int *maxval, int& ominval, int& omaxval, int& numlevels, 
char& coltype, int *spacing, int *transform, char **comment, int& bilevel) {
	char cbuf[12];
	char newline;
	int n_coders;
	fread(cbuf,1,7,cfp);
	cbuf[7] = 0;
	int retval = 0;
	if (strcmp(cbuf,"btpc 5.")) {
		// fprintf(stderr, "This is not a valid btpc 5 file\n");
		return -2;
		}
	fscanf(cfp,"%s %d %d%c",cbuf, &rows,&cols,&newline);
	if (cbuf[0] != 'a') {
		// fprintf(stderr, "This is a version of btpc 5 that I don't ");
		// fprintf(stderr, "recognize. Not going to continue.\n");
		return -3;
		}
	pnmtype = cbuf[1] - '0';// 1 for PGM, 3 for PPM
	numlevels = cbuf[2] - '0';
	n_coders = cbuf[3] - '0';
	coltype = cbuf[4];	// 'G'=GRB, 'Y'=YUV, 'M'=monochrome,
				// 'B'=Bilevel monochrome, 'C'=Bilevel colour
	if (n_coders != numcoders%256) {
		// fprintf(stderr, "Sorry. This picture is coded with ");
		// fprintf(stderr, "%d coders. I can handle ", n_coders);
		// fprintf(stderr, "%d coders only\n",numcoders);
		return -2;
		}
	int i;
	char c;
	for (i = 0; i < 255; i++) {
		c = fgetc(cfp);
		commentbuf[i] = c;
		if (!c)
			break;
		}
	commentbuf[i] = 0;
	if (coltype == 'Y')
		fscanf(cfp,"%d %d %d %d %d %d %d %d %d%c", transform,
		  transform+1, transform+2, transform+3, transform+4,
		  transform+5, transform+6, transform+7, transform+8,
		  &newline);
	if ((coltype == 'B')||(coltype == 'C'))
		bilevel = 1;
	else
		bilevel = 0;
	for (i = 1; i <= numlevels*2; i++)
		spacing[i] = fgetc(cfp);
	for (i = 0; i < pnmtype; i++) {
		minval[i] = fgetc(cfp);
		maxval[i] = fgetc(cfp);
		}
	ominval = fgetc(cfp);
	omaxval = fgetc(cfp);
	return retval;
	}

void to_bilevel(unsigned char ***components, const int pnmtype,
		int *minval, int *maxval, const int& rows, const int& cols) {
	// Make the transformation
	for (int comp = 0; comp < pnmtype; comp++) {
		int thismin = minval[comp];
		int thismax = maxval[comp];
		for (int i = 0; i < rows; i++) {
			unsigned char *p = components[comp][i];
			for (int j = 0; j < cols; j++) {
				if (*p == 0)
					*p = thismin;
				else
					*p = thismax;
				p++;
				}
			}
		}
	}

// Used to have this inline, but the compiler refuses because of all the
// inline code included from the codein[].get()s and codein[].getbit()s.
void unmktopcode(unsigned char topcode,
		int& l0rec, int& l1rec, int& l2rec,
		int& r0rec, int& r1rec, int& r2rec,
		int& cdcoder, int& efcoder,
		rlcomp *codein, const int neg_correct, int pnmtype) {
	int tc = (int) topcode;
	if (pnmtype == 1) {
		if (tc & 0x20) cdcoder = 1;
		if (tc & 0x10) efcoder = 1;
		if (tc & 0xc){	// Just going to test for non-zero
			switch(tc & 0xc) {
				case 0x4:
					l0rec = 1;
					break;
				case 0x8:
					l0rec = neg_correct -1;
					break;
				case 0xc:
					l0rec = codein[0].get();
					if ((l0rec <= 2)&&(codein[0].getbit()))
						l0rec = neg_correct-l0rec;
					break;
				}
			}
		if (tc & 0x3){
			switch(tc & 0x3) {
				case 0x1:
					r0rec = 1;
					break;
				case 0x2:
					r0rec = neg_correct-1;
					break;
				case 0x3:
					r0rec = codein[0].get();
					if ((r0rec <= 2)&&(codein[0].getbit()))
						r0rec = neg_correct-r0rec;
					break;
				}
			}
		}
	else {
		if (tc & 0x80)
			cdcoder = 1;
		if (tc & 0x40) 
			efcoder = 1;
		if (tc & 0x8){
			l0rec = codein[0].get();
			if ((l0rec <= 2)&&(codein[0].getbit()))
				l0rec = neg_correct-l0rec;
			}
		if (tc & 0x10){
			l1rec = codein[1].get();
			if ((l1rec <= 2)&&(codein[0].getbit()))
				l1rec = neg_correct-l1rec;
			}
		if (tc & 0x20){
			l2rec = codein[2].get();
			if ((l2rec <= 2)&&(codein[0].getbit()))
				l2rec = neg_correct-l2rec;
			}
		if (tc & 0x1){
			r0rec = codein[0].get();
			if ((r0rec <= 2)&&(codein[0].getbit()))
				r0rec = neg_correct-r0rec;
			}
		if (tc & 0x2){
			r1rec = codein[1].get();
			if ((r1rec <= 2)&&(codein[0].getbit()))
				r1rec = neg_correct-r1rec;
			}
		if (tc & 0x4){
			r2rec = codein[2].get();
			if ((r2rec <= 2)&&(codein[0].getbit()))
				r2rec = neg_correct-r2rec;
			}
		}
	}
			
void btpc_dec_level(unsigned char ***out, const int& XMAX, const int& YMAX, 
const int& level, const int& sq_spacing, const int& di_spacing,
rlcomp *codein, const int pnmtype, const int& maxval)
{
#ifdef DEBUG
    printf("LEVEL %d\n", level); fflush(stdout);
#endif
    unsigned char *myclip;
    for (int ii = 0; ii < numcoders; ii++)
	codein[ii].readtree();
    prepare_decoder(sq_spacing, di_spacing, maxval);
    int error;
    int X, Y;
    int p0, p1, p2, q0, q1, q2, r0, r1, r2, s0, s1, s2;
    int t0, t1, t2, u0, u1, u2, v0, v1, v2;
    int comp;
    // Surround pels
    int estimate;
    int tempout;
    // int ehigh, elow;

    int sq_ridgethresh = btpc_get_ridgethresh(sq_spacing);
    if (maxval < sq_ridgethresh)	// Use maxval rather than maxval-minval
	sq_ridgethresh = maxval;	// because range is just based on max
    int di_ridgethresh = btpc_get_ridgethresh(di_spacing);
    if (maxval < di_ridgethresh)
	di_ridgethresh = maxval;
    const int XINTERVAL = 1 << level;
    const int YINTERVAL = 1 << level;
    const int XHINT = XINTERVAL >> 1;
    const int YHINT = YINTERVAL >> 1;
    const int XQINT = XHINT >> 1;
    const int YQINT = YHINT >> 1;
    for (Y = YHINT; Y < YMAX; Y += YINTERVAL)
    {   
#ifdef DEBUG
	printf("Y = %d\n",Y); fflush(stdout);
#endif
	int yplus = Y+YHINT;
	int yqplus = Y+YQINT;
	int yqminus = Y-YQINT;
	int yminus = Y-YHINT;
	int y2minus = Y-YINTERVAL;
	if (y2minus < 0)
		y2minus = 0;
	int y3minus = y2minus-YHINT;
	if (y3minus < 0)
		y3minus = 0;

	// All the pointers for three planes
	unsigned char *pyplus0, *pyplus1, *pyplus2;
	unsigned char *pyqplus0, *pyqplus1, *pyqplus2;
	unsigned char *pyqminus0, *pyqminus1, *pyqminus2;
	unsigned char *pyminus0, *pyminus1, *pyminus2;
	unsigned char *py2minus0, *py2minus1, *py2minus2;
	unsigned char *py3minus0, *py3minus1, *py3minus2;
	unsigned char *py0, *py1, *py2;
	if (yplus == YMAX) {
		pyplus0 = out[0][yminus]; // So at bottom of picture, 
			      	    	  // take from previous line
		pyplus1 = out[1][yminus];
		pyplus2 = out[2][yminus];
		}
	else {
		pyplus0 = out[0][yplus];
		pyplus1 = out[1][yplus];
		pyplus2 = out[2][yplus];
		}
	pyqplus0 = out[0][yqplus];
	pyqminus0 = out[0][yqminus];
	pyminus0 = out[0][yminus];
	py2minus0 = out[0][y2minus];
	py3minus0 = out[0][y3minus];
	py0 = out[0][Y];
	pyqplus1 = out[1][yqplus];
	pyqminus1 = out[1][yqminus];
	pyminus1 = out[1][yminus];
	py2minus1 = out[1][y2minus];
	py3minus1 = out[1][y3minus];
	py1 = out[1][Y];
	pyqplus2 = out[2][yqplus];
	pyqminus2 = out[2][yqminus];
	pyminus2 = out[2][yminus];
	py2minus2 = out[2][y2minus];
	py3minus2 = out[2][y3minus];
	py2 = out[2][Y];
	// Set up q and r and v for first column
	v0 = q0 = pyminus0[0];
	r0 = pyplus0[0];
	v1 = q1 = pyminus1[0];
	r1 = pyplus1[0];
	v2 = q2 = pyminus2[0];
	r2 = pyplus2[0];
	unsigned char topcode;
        int leftsibling = 1;
	int xplus = XINTERVAL;
	int xminus = 0;
	for (X = XHINT; X < XMAX; X+=XINTERVAL, xplus+=XINTERVAL, xminus+=XINTERVAL) 
	    {
#ifdef DEBUG
	    if (leftsibling)
	    	printf("X = %d\n",X); fflush(stdout);
#endif
	    p0 = q0; s0 = r0; u0 = v0;
	    p1 = q1; s1 = r1; u1 = v1;
	    p2 = q2; s2 = r2; u2 = v2;
	    if (xplus != XMAX) {
		q0 = pyminus0[xplus]; r0 = pyplus0[xplus];
		q1 = pyminus1[xplus]; r1 = pyplus1[xplus];
		q2 = pyminus2[xplus]; r2 = pyplus2[xplus];
		}
	register int predictor;
	int cdcoder;
	int efcoder;
	int cpropagate, dpropagate;
	int a0rec = 0; int a1rec = 0; int a2rec = 0; 
	int b0rec; int b1rec; int b2rec; 
	int c0rec = 0; int c1rec = 0; int c2rec = 0; 
	int d0rec = 0; int d1rec = 0; int d2rec = 0; 
	topcode = 0;
	if (leftsibling) {
		b0rec = 0; b1rec = 0; b2rec = 0;
		cdcoder = efcoder = 0;
		if (py0[X]) {	// Still some non-zero values below here
			topcode = codein[SST+3].get();
			unmktopcode(topcode,
				a0rec, a1rec, a2rec, b0rec, b1rec, b2rec,
				cdcoder,efcoder,codein+SST,sq_range,
				pnmtype);
			}
		leftsibling = 0;
		}
	else {
		a0rec = b0rec; a1rec = b1rec; a2rec = b2rec;
		cdcoder = efcoder;
		leftsibling = 1;
		}
	// Set up predictor and do prediction in plane 0
	error = a0rec;
	comp = 0;
	myclip = clip[0];
	predictor = 255;	// To indicate first plane
#define SQUARE 1
#define COMP0
#include "macro.cpp"
#undef COMP0
#undef SQUARE
	// If necessary use this predictor in planes 1 and 2
	if (pnmtype > 1) {
		error = a1rec;
		comp = 1;
		myclip = clip[1];
#define SQUARE 1
#define COMP1
#include "macro.cpp"
#undef COMP1
#undef SQUARE
		error = a2rec;
		comp = 2;
		myclip = clip[2];
#define SQUARE 1
#define COMP2
#include "macro.cpp"
#undef COMP2
#undef SQUARE
		}
	// Now onto the diamond bands
	topcode = 0;
	cpropagate = dpropagate = 0;
	if (cdcoder) {
		topcode = codein[DST+3].get();
		unmktopcode(topcode,
			c0rec, c1rec, c2rec, d0rec, d1rec, d2rec,
			cpropagate, dpropagate,codein+DST,di_range,
			pnmtype);
		}

	// Set up predictor and do band 0
	if (!yminus) {
		t0 = p0; t1 = p1; t2 = p2;
		}
	else {
		t0 = py2minus0[X];
		t1 = py2minus1[X];
		t2 = py2minus2[X];
		}
	error = c0rec;
	comp = 0;
	myclip = clip[0];
	predictor = 255;
#define DIAMOND_LEFT 1
#define COMP0
#include "macro.cpp"
#undef COMP0
#undef DIAMOND_LEFT
	// If necessary use this predictor in planes 1 and 2
	if (pnmtype > 1) {
		error = c1rec;
		comp = 1;
		myclip = clip[1];
#define DIAMOND_LEFT 1
#define COMP1
#include "macro.cpp"
#undef COMP1
#undef DIAMOND_LEFT
		error = c2rec;
		comp = 2;
		myclip = clip[2];
#define DIAMOND_LEFT 1
#define COMP2
#include "macro.cpp"
#undef COMP2
#undef DIAMOND_LEFT
		}
	error = d0rec;
	comp = 0;
	myclip = clip[0];
	predictor = 255;
#define DIAMOND_RIGHT 1
#define COMP0
#include "macro.cpp"
#undef COMP0
#undef DIAMOND_RIGHT
	// If necessary use this predictor in planes 1 and 2
	if (pnmtype > 1) {
		error = d1rec;
		comp = 1;
		myclip = clip[1];
#define DIAMOND_RIGHT 1
#define COMP1
#include "macro.cpp"
#undef COMP1
#undef DIAMOND_RIGHT
		error = d2rec;
		comp = 2;
		myclip = clip[2];
#define DIAMOND_RIGHT 1
#define COMP2
#include "macro.cpp"
#undef COMP2
#undef DIAMOND_RIGHT
		}
	}
    }
}

void btpc_dec_plane(unsigned char ***out,const int& xmax, const int& ymax,
    const int& fromlevel, const int *spacing,
    rlcomp *codein, const int& minval, const int& maxval, const int pnmtype) {
	// printf("decplane\n");
	int X, Y;
	int predpel;
	// Fill the first level pels with 1's to indicate that these
	// errors must be fetched (not below the leaf of a tree
	int inc = 1 << fromlevel;
	for (Y = inc/2; Y < ymax; Y += inc)
	for (X = inc/2; X < xmax; X += inc) {
		out[0][Y][X] = 1;
		}
	
	// Get dpcm lowpass
	for (int comp = 0; comp < pnmtype; comp++) {
	    codein[0].readtree();
	    for (Y = 0; Y < ymax; Y += inc)
	    for (X = 0; X < xmax; X += inc) {
		predpel =  (X ? out[comp][Y][X-inc] : 
				(Y ? out[comp][Y-inc][X] : 0));
		int temp = codein[0].get() + predpel;
		if (temp < 0) temp += 256;
		if (temp > 255) temp -= 256;
		out[comp][Y][X] = temp;
		}
	    }
	int tmaxval[3], tminval[3];
	tmaxval[0] = tmaxval[1] = tmaxval[2] = maxval;
	tminval[0] = tminval[1] = tminval[2] = minval;
	initclip(tminval, tmaxval,pnmtype);
	int band = 1;
	for (int level = fromlevel; level > 0; level--, band += 2) {
		btpc_dec_level(out,xmax,ymax,level,spacing[band],
			spacing[band+1], codein, pnmtype, maxval);
		}
	}

static int numlevels;
int btpc_decode(unsigned char ***components, FILE *cfp,
    int& truerows, int& truecols, int& pnmtype, int& topmaxval,
    char **comment) {
/*
	// Use old stdlib for input because it seems to be more efficient
	FILE *cfp = fopen(filename,"rb");
	if (!cfp) {
		return -1;
		}
*/
	int minval[3], maxval[3];// Min and max values of original picture
	int ominval, omaxval; // Overall min and max of transformed picture
	char coltype;
	int spacing[65];
	int i;
	int compnum;
	int transform[9];
	int bilevel;
	if ((i = btpc_read_header(cfp, pnmtype, truerows, truecols,  
	    minval, maxval, ominval, omaxval,
	    numlevels, coltype, spacing, transform, comment, bilevel))) {
		fclose(cfp);
		return i;
		}
	*comment = commentbuf;
	// Make XMAX and YMAX equal to next highest multiple of 16
	int YMAX = ((truerows+(1<<numlevels)-1)>>numlevels)<<numlevels;
	int XMAX = ((truecols+(1<<(numlevels+1))-1)>>(numlevels+1))<<(numlevels+1);
	// Set up decoders
	cfileio fio(cfp);
	rlcomp codein[numcoders];
	for (i = 0; i < numcoders; i++) {
		// codein[i].init();
		codein[i].attach(&fio);
		codein[i].indexnumber(i); // For debugging
#ifdef DEBUG
		codein[i].debugon();
#endif
		}

	// Allocate space for decoded picture
	// If decoding into an existing array (e.g. for immediate display)
	// create line pointers out[i], and point at rows of array.
	unsigned char **out;
	for (compnum = 0; compnum < pnmtype; compnum++)
	    {
	    components[compnum] = out = new unsigned char*[YMAX];
            for (i = 0; i < YMAX; i++) {
                out[i] = new unsigned char[XMAX];
            	memset(out[i],0,XMAX);
		}
	    }
	if (pnmtype == 1) {
		components[2] = components[1] = components[0];
		}
	// Decode all components (CHANGE NAME OF btpc_dec_plane LATER
	btpc_dec_plane(components,XMAX,YMAX,numlevels,
		spacing, codein, ominval, omaxval, pnmtype);
	// Convert colourspace back if necessary
	if (coltype == 'Y') {
		const unsigned char **clip = initclip(minval,maxval,pnmtype);
		// ycc_rgb(components,truerows,truecols,clip,'0');
		pca_rgb(components,transform,truerows,truecols,clip);
		}
	if ((coltype == 'B')||(coltype == 'C'))
		to_bilevel(components, pnmtype, minval, maxval, truerows, 
			truecols);
	topmaxval = maxval[0];
	if (pnmtype == 3) {
		if (maxval[1] > topmaxval)
			topmaxval = maxval[1];
		if (maxval[2] > topmaxval)
			topmaxval = maxval[2];
		}
	fclose(cfp);
	return 1;
	}

void btpc_finish(unsigned char ***components, int truerows, int truecols,
    int pnmtype){
	int YMAX = ((truerows+(1<<numlevels)-1)>>numlevels)<<numlevels;
	for (int compnum = 0; compnum < pnmtype; compnum++) {
		unsigned char **out = components[compnum];
		for (int i = 0; i < YMAX; i++) 
			delete [] out[i];
		delete [] out;
		}
	}
